<?php

/**
 * CacheDecor Component
 * 将一次函数或方法调用按指定参数配置进行缓存
 * 
 * @example
 * //手动调用函数并使用CacheDecor
 * Yii::app()->cacheDecor->run('strval',array('some string'));
 * 
 * //对函数的CacheDecor进行配置
 * function testCacheDecor($arg1,$arg2) {
 * 		static $cacheDecorConfig=array('keys'=>'arg1');
 * 		return $arg1;
 * }
 *	
 * Yii::app()->cacheDecor->run('testCacheDecor',array('arg1','arg2'));//将会自动读取到配置
 * 
 * //在类中对成员方法使用自动CacheDecor
 * //首先实现下面的格式__call方法
 * public function __call($name,$args) {
 * 		$ret=Yii::app()->cacheDecor->call($this,$name,$args);
 * 		if ($ret==CacheDecor::CALL_FAIL) {
 * 			return parent::__call($name,$args); //如果父类有__call方法的话
 * 		} else {
 * 			return $ret;
 * 		}
 * }
 * //类中声明的方法必须为public （除非可以实现__callStatic方法，否则不支持对static方法进行自动缓存），并加上前缀
 * public function cache_deco_methodName() {
 * 		//static $cacheDecorConfig=array();
 * }
 * 
 * //方法调用时不加cache_deco_前缀
 * $this->methodName();
 * 
 * 
 * 
 * 
 * */
class CacheDecor extends CComponent {
	
	
	/**
	 * @const CALL_FAIL 在对象上找不到对应名为cacheMethodName的方法时返回此标识表示调用失败
	 */
	const CALL_FAIL="\x00CacheDecorFlag(CacheDecor->call::Method Call Failed)";
	
	/**
	 * @var array $defaultConfig 外部配置
	 */
	public $defaultConfig=array();

	/**
	 * @var array $_defaultConfig  内部 默认配置项 内部
	 */
	private $_defaultConfig=array(
		/*
		 * 将哪些参数作为缓存的key，存在四种类型取值
		 * 
		 * 指定pattern，pattern至少包含一个“%(参数名称)格式化标识”
		 * 1.'KeyPrefix%(argName)s'  类似于Python中的字符串format，括号内为参数名称，紧跟其后的是将对其如何format的标识（此标识必须是sprintf方法支持的标识，如s表示字符串）
		 * 
		 * 下面三种方式均对传入的值进行序列化，因此值的类型是敏感的，并且它们不能包含%
		 * 2.*（默认），将取函数或方法名称加上所有参数名和传入值生成Key
		 * 3.'param1,param2'，以逗号分隔的参数的名称，这样将根据函数名称加上所列出的参数名称和传入值生成Key
		 * 
		 * 这种情况用于多个函数共享缓存，并且值类型敏感
		 * 4.array('CustomKeyPrefix','param1,param2')，不取函数名称，而取第一个元素CustomKeyPrefix作键前缀，'第二个元素取值同如上2,3两种情况
		 * 
		 * 
		 */
		'keys'=>'*',//默认取所有参数，全部自动生成
		'expire'=>0,//缓存过期时间
		'dependency'=>null,//依赖关系(ICacheDependency的子类的实例)
	);
	
	/**
	 * @var string $methodPrefix 针对需要自动应用Cache类实例方法的所使用的前缀（注意调用这些方法时不需要加此前缀）
	 */
	public $methodPrefix='cache_deco_';
	
	/**
	 * @var string $cacheID 将使用的Cache组件的component ID
	 */
	public $cacheID='cache';
	
	/**
	 * @var string $cacheKeyPrefix 
	 */
	public $cacheKeyPrefix='CacheDecor::';
	
	/**
	 * @var string $configStaticVarName 在函数或方法中的用于配置CacheDecor的static的变量的名字，此static变量值 必须为config数组（可以不写此变量 以使用默认配置）
	 */
	public $configStaticVarName='cacheDecorConfig';

	
	/**
	 * @var bool $enable 标识是否启用缓存
	 */
	public $enable=true;

	private $cache;

	
	public function init() {
		$this->cache=Yii::app()->{$this->cacheID};
		$this->defaultConfig=array_merge($this->_defaultConfig,$this->defaultConfig);
	}
	
	public function setCache(ICache $cache) {
		$this->cache=$cache;
	}
	
	public function getCache() {
		return $this->cache;
	}
	
	
	/**
	 * call 方法，在需要使用自动化方法调用缓存的类实现__call方法，在__call方法中调用此方法
	 * @example 
	 * public function __call($name,$args) {
	 * 		if (Yii::app()->cacheDecor->call($this,$name,$args)===CacheDecor::CALL_FAIL) {
	 * 			//return parent::__call($name,$value); //可以在适当时调用失败再继续调用父类__call
	 * 		}
	 * }
	 * 
	 * @param object $instance 对应的类实例，传入$this
	 * @param string $name 方法名称
	 * @param array  $args传入的参数数组
	 * @return mixed 对应方法的返回值，如果方法不存在，则返回CALL_FAIL常量标识
	 */
	public function call($instance,$name,array $args=array()) {
		$methodName=$this->methodPrefix.$name;
		if (method_exists($instance,$methodName)) {
			return $this->run(array($instance,$methodName),$args);
		} else {
			return self::CALL_FAIL;
		}
			
	}
	
	/**
	 * Run a function or method with Cache Decoration
	 * 注意，本方法不会去检测callback是否可调用
	 * @example 
	 * $cacheDecor->run('strval',array('I am string!'))
	 * //methodTest方法为$this对象的方法（这里不会自动加前缀的）,keys指出只以名为arg1的参数为CacheKey，不考虑第二个参数
	 * $cacheDecor->run(array($this,'methodTest'),array('arg1','arg2'),array('keys'=>'arg1'))
	 * 
	 * @param mixed $callback must be a valid callback(a function name or array(ClassName,'staticMethod') or array($instance,'methodName') or lambda)
	 * @param array $args is the parameters that passed to the function when call it
	 * @param array $config is an config array that contains cache mode(Cache return or output) and params used as cache keys
	 * @options $config
	 * array(
	 * 		//参见defaultConfig注释
	 *		'keys'=>'*',//将哪些参数作为缓存的key(以逗号分隔的参数名称),默认将所有参数当成变因作为缓存的键，参数值将进行serialize
	 * )
	 * 
	 * @return mixed   返回值为$callback对应的返回
	 */
	public function run($callback,array $inputArgs=array(),array $config=array()) {
		if (!$this->enable) {//enable配置为false,将直接调用而不使用缓存
			return call_user_func_array($callback,$inputArgs);
		}

		$reflector=$this->createReflection($callback);//创建反射对象
		$config=$this->mergeConfig($reflector,$config);//获取并合并配置

		//将数字下标的args转换成以对应参数名为键的关联数组
		$argsMap=$this->fillArguments($reflector->getParameters(),$inputArgs);
		
		$key=$this->genKey($config['keys'],$argsMap);//生成对应于$args和$keys配置一致的Key
		$cache=$this->cache;
		$ret=$cache->get($key);
		if (false===$ret) {//缓存中没有数据，则执行对应方法并设置缓存
			$ret=call_user_func_array($callback,$inputArgs);
			$cache->set($key,$ret,$config['expire'],$config['dependency']);
		}
		return $ret;
		
	}
	
	
	/**
	 * genKey 
	 * 
	 * @param array/string $keys  格式为array('CustomKeyPrefix','arg1,arg2')或'Autolabel_%(id)s'的keys配置项
	 * @param array $args  参数名=>参数值  对的关联数组
	 * @return string CacheKey
	 */
	public function genKey($keys,array $args=array()) {
		
		if (is_array($keys)) {
			//根据keys配置和传入参数值的不同生成不同的args key(args key中为传入参数名称对数组序列化)
			$key=$this->genSerialKey($keys,$args);
		} else {//自定义pattern形式的key			
			//将原来的$args数字下标数组转换成以参数名为键的关联数组
			$key=$this->genStrKey($keys,$args);
		}
		return $this->cacheKeyPrefix.$key;
		
	}
	
	
	/**
	 * 
	 * @param string $keys 选择性要取的参数
	 * @param array $args 实参
	 */
	public function genSerialKey($keys,$args) {
		if ($keys[1]!=='*') {
			$unsets=array_diff(array_keys($args),explode(',',$keys[1]));
			foreach($unsets as $index) {
				unset($args[$index]);
			}
		}
		return $keys[0].md5(serialize($args));
	}
	
	/**
	 * genStrKey 根据pattern(字符串format)和传入参数args(关联数组)生成一个CacheKey
	 * 
	 * @param string $pattern format字符串
	 * @param array $args 包含参数的关联数组
	 * @return string 生成的Key
	 */
	public function genStrKey($pattern,array $args) {
		return $this->format($pattern,$args);	
	}
	
	
	private function fillArguments($params,$args) {
		$data=array();
		foreach($params as $p) {
			$pos=$p->getPosition();
			$name=$p->getName();
			$data[$name]=isset($args[$pos])?$args[$pos]:($p->isDefaultValueAvailable()?$p->getDefaultValue():NULL);
		}
		return $data;
	}
	

	
	/**
	 * createReflection 根据$callback返回对应方法或函数的Reflector
	 * 
	 * @access private
	 * @return $reflector
	 */
	private function createReflection($callback) {
		if (is_string($callback) && strpos($callback,'::')!==false) {//静态方法调用  "Class::method"
			$callback=explode('::',$callback,2);
		}
		if (is_array($callback)) {
			list($cls,$method)=$callback;
			if (!is_string($cls)) $cls=get_class($cls);//实例
			$ref=new ReflectionMethod($cls,$method);
		} else {
			$ref=new ReflectionFunction($callback);
		}
		return $ref;
	}
	
	
	/**
	 * getCacheConfig 获取一个函数的静态变量中的CacheDecor配置
	 * 
	 * @access private
	 * @param Reflector $reflector 一个函数或方法的Reflection对象
	 * @param array $config 
	 * @return array $config数组
	 */
	private function mergeConfig(Reflector $reflector,array $config=array()) {
		$vars=$reflector->getStaticVariables();
		$decorConfig=array();
		if (isset($vars[$this->configStaticVarName]) && is_array($vars[$this->configStaticVarName])) {
			$decorConfig=$vars[$this->configStaticVarName];			
		}
		$config=array_merge($this->defaultConfig,$decorConfig,$config);
		if ($this->isAutoPrefixKeys($config['keys'])) {
			$callbackPrefix=sprintf('%s::%s::%s',get_class($reflector),isset($reflector->class)?$reflector->class:'Function',$reflector->name);
			$config['keys']=array($callbackPrefix,$config['keys']);
		}
		return $config;
		
	}
	
	private function isAutoPrefixKeys($keys) {
		return  (is_string($keys) && strpos($keys,'%(')===false);
	}
	
	
	/**
	 * 
	 * @deprecated
	 */
	private function _format($pattern,array $args) {
		foreach ($args as $k=>$v) {
			$pkey="%($k)";
			$pos=strpos($pattern,$pkey);
			if ($pos===false) continue;
			$tag=$pattern{$pos+strlen($pkey)};
			$pattern=str_replace($pkey.$tag,sprintf("%{$tag}",$v),$pattern);
		}
		return $pattern;
	}
	
	//Test
	public function format($pattern,array $args) {
		$values=array();
		foreach ($args as $k=>$v) {
			$pkey="%($k)";
			$pos=strpos($pattern,$pkey);
			if ($pos===false) continue;
			$values[$pos]=$v;
			$pattern=str_replace($pkey,'%',$pattern);
		}
		$values=array_values($values);
		array_unshift($values,$pattern);

		return call_user_func_array('sprintf',$values);
	}
	
}
